import tempfile

import numpy as np

from gtnet.sequence import FastaSequenceEncoder, FastaReader

VOCAB = 'ACYWSKDVNTGRMHB'


def get_fasta():
    """TODO - modify this to write to a temp file and return the tempfile path"""
    fl = [">seq1 Sulfolobus",
          "AAGCTNGGGCATTTCAGGGTGAGCCCGGGCAATACAGGG",
          ">seq2 Halobacteria",
          "AAGCCTTGGCAGTGCAGGGTGAGCCGTGGCCGGGCACGGTA",
          ">seq3 Salmonella",
          "ACCGGTTGGCCGTTCAGGGTACAGGTTGGCCGTTCAGGGTAA",
          ">seq4 E.coli",
          "AAACCCTTGCCGTTACGCTTAAACCGAGGCCGGGACAC",
          ">seq5 Y.pestis",
          "AAACCCTTGCCGGTACGCTTAAACCATTGCCGGTACGCTT"]
    _, path = tempfile.mkstemp()
    with open(path, 'w') as f:
        for line in fl:
            print(line, file=f)
    return path


def test_encoder():
    encoder = FastaSequenceEncoder(3, 2, vocab=VOCAB, min_seq_len=1)
    seq = np.array(list("ACTGNGTCNA"))
    batch = encoder.encode(seq)
    expected = np.array([
        [[  0,  1,  9],   # ACT
         [  9, 10,  8],   # TGN
         [  8, 10,  9],   # NGT
         [  9,  1,  8],   # TCN
         [  8,  0,  8]],  # NA

        [[  9,  8, 10],   # TNG
         [ 10,  0,  1],   # GAC
         [  1,  8,  1],   # CNC
         [  1,  0, 10],   # CAG
         [ 10,  9,  8]]   # GT

    ])
    np.testing.assert_array_equal(expected, batch)


def test_reader():
    encoder = FastaSequenceEncoder(8, 2, vocab=VOCAB, min_seq_len=4)
    fa = get_fasta()
    reader = FastaReader(encoder, fa)
    i = 0
    for seq in reader:
        assert len(seq) == 4
        i += 1
    assert i == 5




import hashlib
from importlib.resources import files
import json
import logging
import os
import sys
import urllib
import warnings
import zipfile

import numpy as np
import torch
import torch.nn as nn


def parse_logger(string):
    if not string:
        ret = logging.getLogger()
        hdlr = logging.StreamHandler(sys.stderr)
    else:
        ret = logging.getLogger(string)
        hdlr = logging.FileHandler(string)
    ret.setLevel(logging.INFO)
    ret.addHandler(hdlr)
    hdlr.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    return ret


def get_logger():
    return parse_logger('')


class DeployPkg:
    """A class to handle loading and manipulating the deployment package"""

    _deploy_pkg_url = "https://osf.io/download/qf46x/"

    _checksum = "623aa991fb0d74e874b7d0da25496c26"

    _manifest_name = 'manifest.json'

    @classmethod
    def check_pkg(cls):
        deploy_dir = files(__package__).joinpath('deploy_pkg')
        total = os.path.getsize(os.path.join(deploy_dir, cls._manifest_name))
        if total == 0:
            msg = ("Downloading GTNet deployment package. This will only happen on the first invocation "
                   "of gtnet predict or gtnet classify")
            warnings.warn(msg)
            zip_path = files(__package__).joinpath('deploy_pkg.zip')
            urllib.request.urlretrieve(cls._deploy_pkg_url, zip_path)
            dl_checksum = hashlib.md5(open(zip_path,'rb').read()).hexdigest()
            if dl_checksum != cls._checksum:
                raise ValueError(f"Downloaded deployment package {zip_path} does not match expected checksum")
            with zipfile.ZipFile(zip_path) as zip_ref:
                zip_ref.extractall(os.path.dirname(deploy_dir))
            os.remove(zip_path)
        return deploy_dir


    def __init__(self):
        self.deploy_dir = self.check_pkg()

        self._manifest = None

    def path(self, path):
        """Map paths to be relative to current working directory"""
        return os.path.join(self.deploy_dir, path)

    @property
    def manifest(self):
        if self._manifest is None:
            with open(self.path(self._manifest_name), 'r') as f:
                self._manifest = json.load(f)
        return self._manifest

    def __getitem__(self, key):
        return self.manifest[key]

    def __setitem__(self, key, val):
        self.manifest[key] = val


def load_deploy_pkg(for_predict=False, for_filter=False, contigs=False):
    if not (for_predict or for_filter):
        for_predict = True
        for_filter = True

    pkg = DeployPkg()
    key = 'contigs' if contigs else 'bins'

    ret = list()
    if for_predict:
        tmp_conf_model = dict()
        for cm_data, taxa_data in zip(pkg['conf_model'][key], pkg['taxa']):
            if cm_data['level'] != taxa_data['level']:
                raise ValueError("Taxonomic levels are out of order in manifest file")
            cm_data['taxa'] = np.array(taxa_data['taxa'])

            cm_data['model'] = torch.jit.load(pkg.path(cm_data.pop('model')))

            tmp_conf_model[cm_data['level']] = cm_data

        ret.append(torch.jit.load(pkg.path(pkg['nn_model'])))
        ret.append(tmp_conf_model)
        ret.append(pkg['training_config'])
        ret.append("".join(pkg['vocabulary']))

    if for_filter:
        tmp_roc = dict()
        for cm_data in pkg['conf_model'][key]:
            tmp_roc[cm_data['level']] = np.load(pkg.path(cm_data['roc']))
        ret.append(tmp_roc)

    return tuple(ret) if len(ret) > 1 else ret[0]


class GPUModel(nn.Module):

    def __init__(self, model, device):
        super().__init__()
        self.device = device
        self.model = model.to(self.device)

    def forward(self, x):
        x = x.to(self.device)
        return self.model(x).cpu()


def check_cuda(parser):
    if torch.cuda.is_available():
        parser.add_argument('-g', '--gpu', action='store_true', default=False, help='Use GPU')
        parser.add_argument('-D', '--device_id', type=int, default=0,
                            choices=torch.arange(torch.cuda.device_count()).tolist(),
                            help='the device ID of the GPU to use')


def check_device(args):
    if getattr(args, 'gpu', False):
        return torch.device(args.device_id)
    return torch.device('cpu')


def write_csv(output, args):

    # write out data
    if args.output is None:
        outf = sys.stdout
    else:
        outf = open(args.output, 'w')
    output.to_csv(outf, index=True)



import io
import sys
from importlib import import_module


class Command:
    def __init__(self, module, doc):
        ar = f'gtnet.{module}'.split('.')
        self.pkg = '.'.join(ar[:-1])
        self.func = ar[-1]
        self.doc = doc

    def get_func(self):
        return getattr(import_module(self.pkg), self.func)


command_dict = {
    "classify": Command("classify.classify",
                        "Perform taxonomic classificaiton for all sequences"),
    "predict":  Command("predict.predict",
                        "Predict taxonomy of all sequences at each taxonomic level"),
    "filter":   Command("filter.filter",
                        "Filter taxonomic classifications from the predict command"),
    "test":     Command("run.run_test", "Run gtnet on a sample dataset provided"),
}


def print_help():
    sio = io.StringIO()
    print('Usage: gtnet <command> [options]', file=sio)
    print('Available commands are:\n', file=sio)
    for c, f in command_dict.items():
        nspaces = 16 - len(c)
        print(f'    {c}' + ' '*nspaces + f.doc, file=sio)
    print('    help            print this usage statement\n', file=sio)
    sys.stdout.write(sio.getvalue())


def run():
    if len(sys.argv) == 1 or sys.argv[1] in ('help', '-h', '--help'):
        print_help()
    else:
        cmd = sys.argv[1]
        func = command_dict[cmd].get_func()
        func(sys.argv[2:])

if __name__ == '__main__':
    run()



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("package-name")
except PackageNotFoundError:
    # package is not installed
    __version__ = '0.0.1'
    pass

del version, PackageNotFoundError



import argparse
import logging
from time import time

from .utils import check_cuda, check_device, get_logger, load_deploy_pkg, write_csv
from .filter import filter_predictions, get_cutoffs
from .predict import run_torchscript_inference


DEFAULT_N_CHUNKS = 10000


def classify(argv=None):
    """
    Get taxonomic classification for each sequence in a Fasta file.

    Parameters
    ----------

    argv : Namespace, default=sys.argv
        The command-line arguments to use for running this command
    """
    desc = "Get filtered taxonomic classification for each sequence in a Fasta file."
    epi = ("This command will output a taxonomic classification filtered to a specified false-positive rate "
           "for each sequence")

    parser = argparse.ArgumentParser(description=desc, epilog=epi)
    parser.add_argument('fastas', type=str, nargs='+', help='the Fasta files to do taxonomic classification on')
    parser.add_argument('-s', '--seqs', action='store_true', help='provide classification for sequences')
    parser.add_argument('-c', '--n_chunks', type=int, default=DEFAULT_N_CHUNKS,
                        help='the number of sequence chunks to process at a time')
    parser.add_argument('-o', '--output', type=str, default=None, help='the output file to save classifications to')
    check_cuda(parser)
    parser.add_argument('-f', '--fpr', default=0.05, type=float, help='the false-positive rate to classify to')
    parser.add_argument('-d', '--debug', action='store_true', default=False,
                        help='print specific information about sequences')

    args = parser.parse_args(argv)

    before = time()

    logger = get_logger()
    if args.debug:
        logger.setLevel(logging.DEBUG)

    device = check_device(args)

    model, conf_models, train_conf, vocab, rocs = load_deploy_pkg(for_predict=True, for_filter=True, contigs=args.seqs)

    window = train_conf['window']
    step = train_conf['step']

    logger.info(f'Getting class predictions for each contig in {",".join(args.fastas)}')
    output = run_torchscript_inference(args.fastas, model, conf_models, window, step, vocab, seqs=args.seqs,
                                      device=device, logger=logger)

    logger.info(f'Getting probability cutoffs for target false-positive rate of {args.fpr}')
    cutoffs = get_cutoffs(rocs, args.fpr)

    logger.info('Filtering class predictions')
    output = filter_predictions(output, cutoffs)

    write_csv(output, args)

    after = time()
    logger.info(f'Took {after - before:.1f} seconds')



import argparse
from collections import Counter
import logging
from time import time

import pandas as pd
import torch

from .sequence import FastaReader, FastaSequenceEncoder
from .utils import check_cuda, check_device, get_logger, load_deploy_pkg, write_csv


DEFAULT_N_CHUNKS = 10000


def predict(argv=None):
    """
    Get network predictions for each sequence in Fasta file

    Parameters
    ----------

    argv : Namespace, default=sys.argv
        The command-line arguments to use for running this command
    """
    desc = "Get network predictions for each sequence in a Fasta file"
    epi = ("This command will output the best classification for every sequence at every taxonomic level, including "
           "confidence scores for each taxonomic level. These classifications can be subsequently filtered using the "
           "'filter' command. See the 'classify' command for getting pre-filtered classifications")

    parser = argparse.ArgumentParser(description=desc, epilog=epi)
    parser.add_argument('fastas', type=str, nargs='+', help='the Fasta files to do taxonomic classification on')
    parser.add_argument('-s', '--seqs', action='store_true', help='provide classification for sequences')
    parser.add_argument('-c', '--n_chunks', type=int, default=DEFAULT_N_CHUNKS,
                        help='the number of sequence chunks to process at a time')
    parser.add_argument('-o', '--output', type=str, default=None, help='the output file to save classifications to')
    check_cuda(parser)
    parser.add_argument('-d', '--debug', action='store_true', default=False,
                        help='print specific information about sequences')

    args = parser.parse_args(argv)

    before = time()

    logger = get_logger()
    if args.debug:
        logger.setLevel(logging.DEBUG)

    device = check_device(args)

    model, conf_models, train_conf, vocab = load_deploy_pkg(for_predict=True, contigs=args.seqs)

    window = train_conf['window']
    step = train_conf['step']

    output = run_torchscript_inference(args.fastas, model, conf_models, window, step, vocab, seqs=args.seqs,
                                       device=device, logger=logger)

    # write out data
    write_csv(output, args)

    after = time()
    logger.info(f'Took {after - before:.1f} seconds')


def run_torchscript_inference(fastas, model, conf_models, window, step, vocab, seqs=False, n_chunks=DEFAULT_N_CHUNKS,
                              device=torch.device('cpu'), logger=None):
    """Run Torchscript inference

    Parameters
    ----------

    fastas : str
        The path to the Fasta file with sequences to do inference on

    model : RecursiveScriptModule
        The Torchscript model to run inference with

    conf_models : dict
        A dictionary with the confidence model for each taxonomic level. Each model should be a RecursiveScriptModule.
        The expected keys in this dict are 'domain', 'phylum', 'class', 'order', 'family', 'genus' and 'species'.

    window : int
        The length of the sliding window to use for doing inference

    step : int
        The length of the step of the sliding window to use for doing inference

    vocab : str
        The vocabulary used for training `model`

    n_chunks : int, default=10000
        The length of the step of the sliding window to use for doing inference

    device : device, default=torch.device('cpu')
        The Pytorch device to run inference on

    logger : Logger
        The Python logger to use when running inference
    """

    if logger is None:
        logger = get_logger()
        logger.setLevel(logging.CRITICAL)

    encoder = FastaSequenceEncoder(window, step, vocab=vocab, device=device)
    reader = FastaReader(encoder, *fastas)

    model = model.to(device)

    output_size = sum(len(lvl['taxa']) for lvl in conf_models.values())

    seqnames = list()
    lengths = list()
    total_chunks = list()
    filepaths = list()
    aggregated = list()

    torch.set_grad_enabled(False)


    logger.info(f'Calculating classifications for all sequences in {", ".join(fastas)}')
    for file_path, seq_name, seq_len, seq_chunks in reader:
        seqnames.append(seq_name)
        lengths.append(seq_len)
        filepaths.append(file_path)

        logger.debug((f'Getting predictions for windows of {seqnames[-1]}, '
                      f'{seq_chunks.shape[1] * 2} chunks, {lengths[-1]} bases'))
        outputs = torch.zeros(output_size, device=device)   # the output from the network for a single sequence
        # sum network outputs from all chunks
        for s in range(0, seq_chunks.shape[1], n_chunks):
            e = s + n_chunks
            outputs += model(seq_chunks[0, s:e]).sum(dim=0)
            outputs += model(seq_chunks[1, s:e]).sum(dim=0)
        # divide by the number of seq_chunks we processed to get a mean output
        aggregated.append(outputs)
        total_chunks.append(seq_chunks.shape[1] * 2.)

        del seq_chunks

    total_chunks = torch.tensor(total_chunks, device=device)

    # aggregate everything we just pulled from the fasta file
    all_levels_aggregated = torch.row_stack(aggregated)
    del aggregated

    if not seqs:
        logger.info('Calculating classifications for bins')

        ctr = Counter(filepaths)
        n_ctgs = list(ctr.values())
        filepaths = list(ctr.keys())

        max_len = list()
        l50 = list()
        lengths = torch.tensor(lengths)

        tmp_chunks = list()
        tmp_aggregated = list()

        s = 0
        for n in n_ctgs:
            e = s + n
            tmp_lens = lengths[s:e].sort(descending=True).values
            max_len.append(tmp_lens[0])
            csum = torch.cumsum(tmp_lens, 0)
            l50.append(tmp_lens[torch.where(csum > (csum[-1] * 0.5))[0][0]])
            tmp_aggregated.append(all_levels_aggregated[s:e].sum(axis=0))
            tmp_chunks.append(total_chunks[s:e].sum())
            s = e

        del total_chunks
        total_chunks = torch.tensor(tmp_chunks, device=device)
        del tmp_chunks

        del all_levels_aggregated
        all_levels_aggregated = torch.row_stack(tmp_aggregated)
        del tmp_aggregated

        features = torch.tensor([n_ctgs, l50, max_len], device=device).T
        output_data = {'file': filepaths}
    else:
        features = torch.tensor(lengths, device=device)[:, None]
        output_data = {'file': filepaths, 'ID': seqnames}


    all_levels_aggregated = ((all_levels_aggregated.T) / total_chunks).T
    indices = list(output_data.keys())

    s = 0
    for lvl, e in zip(model.levels, model.parse):
        conf_model_info = conf_models[lvl]
        taxa = conf_model_info['taxa']
        conf_model = conf_model_info['model'].to(device)

        # determine the number of top k probabilities
        # to use for confidence scoring
        top_k = 2

        logger.debug(f'Getting {lvl} predictions for all sequences')
        aggregated = all_levels_aggregated[:, s:e]
        output_data[lvl] = taxa[torch.argmax(aggregated, dim=1).cpu()]

        # get prediction and maximum probabilities for confidence scoring
        logger.debug(f'Getting max {top_k} probabilities for {lvl}')
        maxprobs = torch.topk(aggregated, top_k, dim=1, largest=True, sorted=True).values

        # build input matrix for confidence model
        logger.debug('Calculating confidence probabilities')
        conf_input = torch.column_stack([features, maxprobs])

        output_data[f'{lvl}_prob'] = conf_model(conf_input).cpu().numpy().squeeze()

        # set next left bound for all_levels_aggregated
        s = e

    output = pd.DataFrame(output_data).set_index(indices)

    return output



from itertools import chain
import logging

import numpy as np
import torch
import torch.nn.functional as F


import torch.multiprocessing as mp
try:
    mp.set_start_method('spawn')
except RuntimeError:
    pass

__all__ = ['FastaSequenceEncoder', 'FastaReader']


class FastaSequenceEncoder:

    def __init__(self, window, step, vocab=None, padval=None, min_seq_len=100, device=torch.device('cpu')):
        self.window = window
        self.step = step
        self.vocab, self.basemap, self.rcmap = self.get_dna_map(vocab=vocab)
        if padval is None:
            self.padval = next(i for i in range(len(self.vocab)) if self.vocab[i].upper() == 'N')
        else:
            self.padval = padval

        if min_seq_len > window:
            raise ValueError("min_seq_len must be less than or equal to window")

        self.padval = self.padval
        self.min_seq_len = min_seq_len
        self.device = device

    def encode(self, seq):
        if len(seq) < self.min_seq_len:
            raise ValueError(f"Minimum sequence length is {self.min_seq_len} - got {len(seq)}")
        if seq.dtype == np.dtype('S1'):
            seq = seq.view(np.uint8)
        elif seq.dtype == np.dtype('U1'):
            seq = seq.astype('S').view(np.uint8)
        elif not issubclass(seq.dtype.type, np.integer):
            raise ValueError('seq must be bytes or integer type')

        fwd = self.basemap[seq]
        rev = self.rcmap[fwd[::-1]]
        fwd = torch.from_numpy(fwd)
        rev = torch.from_numpy(rev)

        C_min = len(seq) % self.step
        if C_min == 0:
            C_min = self.step
        n_short_C = max(((self.min_seq_len - C_min - 1) // self.step) + 1, 0)
        n_C = (len(seq) - 1) // self.step + 1
        n_chunks = n_C - n_short_C
        padlen =  (n_chunks - 1) * self.step + self.window - len(seq)

        fwd = F.pad(fwd, (0, padlen), "constant", self.padval)
        rev = F.pad(rev, (0, padlen), "constant", self.padval)

        ret = torch.row_stack([fwd, rev]).to(self.device)
        ret = ret.unfold(1, self.window, self.step)

        return ret

    @classmethod
    def get_dna_map(cls, vocab=None):
        '''
        Create data structures for mapping DNA sequence to

        Returns
            vocab:      the DNA vocabulary used for building the data structures
            basemap:    a 128 element array for mapping ASCII character values to encoded values
            rcmap:      an array for mapping between complementary characters of encoded values
        '''
        if vocab is None:
            vocab = ('ACYWSKDVNTGRMHB')
        rcmap = cls.get_revcomp_map(vocab)
        basemap = np.zeros(128, dtype=np.uint8)
        # reverse so we store the lowest for self-complementary codes
        for i, c in reversed(list(enumerate(vocab))):
            basemap[ord(c)] = i
            basemap[ord(c.lower())] = i
        basemap[ord('x')] = basemap[ord('X')] = basemap[ord('n')]
        return vocab, basemap, rcmap

    @classmethod
    def get_revcomp_map(cls, vocab):
        chars = {
            'A': 'T',
            'G': 'C',
            'C': 'G',
            'T': 'A',
            'Y': 'R',
            'R': 'Y',
            'W': 'W',
            'S': 'S',
            'K': 'M',
            'M': 'K',
            'D': 'H',
            'V': 'B',
            'H': 'D',
            'B': 'V',
            'X': 'X',
            'N': 'N',
        }

        d = {c: i for i, c in enumerate(vocab)}
        rcmap = np.zeros(len(vocab), dtype=np.uint8)
        for i, base in enumerate(vocab):
            rc_base = chars[base]
            base_i = d[base]
            rc_base_i = d[rc_base]
            rcmap[base_i] = rc_base_i
        return rcmap


class Loader:

    def __init__(self, encoder, *fastas):
        self.encoder = encoder
        self.fastas = fastas

    @staticmethod
    def readfile(fasta_path):
        chars = None
        seqid = None
        with open(fasta_path, 'rb') as f:
            while True:
                line = f.readline()
                if not line:
                    yield seqid, np.fromiter(chain(*chars), dtype=np.uint8)
                    break
                if line[0] == 62:
                    if seqid is not None:
                        yield seqid, np.fromiter(chain(*chars), dtype=np.uint8)
                    seqid = line[1:line.find(32)].decode('utf8')
                    chars = list()
                else:
                    chars.append(line[:-1])

    @classmethod
    def readfiles(cls, encoder, fastas):
        for fa in fastas:
            logging.debug(f'loading {fa}')
            for seqid, values in cls.readfile(fa):
                if len(values) < encoder.min_seq_len:
                    logging.warning((f"Skipping {seqid} from {fa} - length less than "
                                    "minimum sequence length {encoder.min_seq}"))
                    yield (fa, seqid, len(values), torch.zeros((0, 0, 0), dtype=torch.uint8))
                else:
                    batches = encoder.encode(values)
                    val = (fa, seqid, len(values), batches)
                    yield val


class SerialLoader(Loader):

    def __iter__(self):
        return self.readfiles(self.encoder, self.fastas)


class ParallelLoader(Loader):

    _done = 0xDEADBEEF

    @classmethod
    def _target(cls, q, fastas, encoder, finished):
        for val in cls.readfiles(encoder, fastas):
            q.put(val)
        q.put(cls._done)
        finished.wait()

    def __iter__(self):
        self._q = mp.Queue()
        self._finished = mp.Event()
        self._proc = mp.Process(target=self._target, args=(self._q, self.fastas, self.encoder, self._finished))
        self._proc.start()
        return self

    def __next__(self):
        seq = None
        while True:
            seq = self._q.get()
            if seq == self._done:
                self._finished.set()
                self._proc.join()
                raise StopIteration
            break
        return seq


class FastaReader(mp.Process):


    def __init__(self, encoder, *fastas, parallel=False):
        if parallel:
            self.loader = ParallelLoader(encoder, *fastas)
        else:
            self.loader = SerialLoader(encoder, *fastas)

    def __iter__(self):
        return iter(self.loader)
